"""Functions for creating data in numpy."""

from __future__ import annotations

from typing import Any

from jumpy.core import which_dtype, which_np
from jumpy.numpy._types import ndarray


def arange(start: int, stop: int, dtype=None) -> ndarray:
    """Return evenly spaced values within a given interval."""
    if dtype is None:
        return which_np(start, stop).arange(start, stop, dtype=dtype)
    else:
        return which_dtype(dtype).arange(start, stop, dtype=dtype)


def array(object: Any, dtype=None) -> ndarray:
    """Creates an array given a list."""
    if dtype is None:
        try:
            np = which_np(*object)
        except TypeError:
            np = which_np(object)  # object is not iterable (e.g. primitive type)
    else:
        np = which_dtype(dtype)

    return np.array(object, dtype)


def atleast_1d(*arys) -> ndarray:
    """Ensure arrays are all at least 1d (dimensions added to beginning)."""
    return which_np(*arys).atleast_1d(*arys)


def atleast_2d(*arys) -> ndarray:
    """Ensure arrays are all at least 2d (dimensions added to beginning)."""
    return which_np(*arys).atleast_2d(*arys)


def atleast_3d(*arys) -> ndarray:
    """Ensure arrays are all at least 3d (dimensions added to beginning)."""
    return which_np(*arys).atleast_3d(*arys)


def eye(n: int, dtype=float) -> ndarray:
    """Return a 2-D array with ones on the diagonal and zeros elsewhere."""
    return which_dtype(dtype).eye(n, dtype=dtype)


def linspace(start: ndarray, stop: ndarray, num: int, dtype=None) -> ndarray:
    """Return evenly spaced `num` values between `start` and `stop`."""
    if dtype is None:
        return which_np(start, stop, num).linspace(start, stop, num)
    else:
        return which_dtype(dtype).linspace(start, stop, num, dtype=dtype)


def meshgrid(
    *xi, copy: bool = True, sparse: bool = False, indexing: str = "xy"
) -> ndarray:
    """Create N-D coordinate matrices from 1D coordinate vectors."""
    return which_np(xi[0]).meshgrid(*xi, copy=copy, sparse=sparse, indexing=indexing)


def ones(shape, dtype=float) -> ndarray:
    """Return a new array of given shape and type, filled with ones."""
    return which_dtype(dtype).ones(shape, dtype=dtype)


def ones_like(a: ndarray) -> ndarray:
    """Return an array of ones with the same shape and type as a given array."""
    return which_np(a).ones_like(a)


def reshape(a: ndarray, newshape: tuple[int, ...] | int) -> ndarray:
    """Gives a new shape to an array without changing its data."""
    return which_np(a).reshape(a, newshape)


def zeros(shape, dtype=float) -> ndarray:
    """Return a new array of given shape and type, filled with zeros."""
    return which_dtype(dtype).zeros(shape, dtype=dtype)


def zeros_like(a: ndarray) -> ndarray:
    """Return an array of zeros with the same shape and type as a given array."""
    return which_np(a).zeros_like(a)



"""Functions from the `jax.ops` module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as onp

from jumpy.core import which_np

if TYPE_CHECKING:
    from jumpy.numpy import ndarray

try:
    import jax
    import jax.numpy as jnp
except ImportError:
    jax, jnp = None, None


__all__ = ["segment_sum"]


def segment_sum(
    data: ndarray, segment_ids: ndarray, num_segments: int | None = None
) -> ndarray:
    """Computes the sum within segments of an array."""
    if which_np(data, segment_ids) is jnp:
        s = jax.ops.segment_sum(data, segment_ids, num_segments)
    else:
        if num_segments is None:
            num_segments = onp.amax(segment_ids) + 1
        s = onp.zeros((num_segments,) + data.shape[1:])
        onp.add.at(s, segment_ids, data)
    return s



"""Functions from the `jax.lax` module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Sequence, TypeVar

import numpy as onp

from jumpy import is_jax_installed
from jumpy.core import is_jitted, which_np

if TYPE_CHECKING:
    from jumpy.numpy import ndarray

try:
    import jax
    import jax.numpy as jnp
except ImportError:
    jax, jnp = None, None


__all__ = [
    "cond",
    "fori_loop",
    "switch",
    "scan",
    "stop_gradient",
    "top_k",
    "while_loop",
]

Carry = TypeVar("Carry")
X = TypeVar("X")
Y = TypeVar("Y")


def cond(
    pred, true_fun: Callable[..., bool], false_fun: Callable[..., bool], *operands: Any
):
    """Conditionally apply true_fun or false_fun to operands."""
    if is_jitted():
        return jax.lax.cond(pred, true_fun, false_fun, *operands)
    else:
        if pred:
            return true_fun(*operands)
        else:
            return false_fun(*operands)


def switch(index: int, branches: Sequence[Callable], *operands: Any):
    """Conditionally apply exactly one of ``branches`` given by ``index`` operands.

    Has the semantics of the following Python::

        def switch(index, branches, *operands):
          index = clamp(0, index, len(branches) - 1)
          return branches[index](*operands)
    """
    if is_jitted():
        return jax.lax.switch(index, branches, *operands)
    else:
        # NOTE: should we clamp to mimic jax behavior? Or raise an error for out of bounds indices?
        # index = onp.clip(index, 0, len(branches) - 1)
        return branches[index](*operands)


def fori_loop(
    lower: int, upper: int, body_fun: Callable[[int, X], X], init_val: X
) -> X:
    """Call body_fun over range from lower to upper, starting with init_val."""
    if is_jitted():
        return jax.lax.fori_loop(lower, upper, body_fun, init_val)
    else:
        val = init_val
        for i in range(lower, upper):
            val = body_fun(i, val)
        return val


def scan(
    f: Callable[[Carry, X], tuple[Carry, Y]],
    init: Carry,
    xs: X,
    length: int | None = None,
    reverse: bool = False,
    unroll: int = 1,
) -> tuple[Carry, Y]:
    """Scan a function over leading array axes while carrying along state."""
    if not is_jax_installed:
        raise NotImplementedError("This function requires the jax module")

    if is_jitted():
        return jax.lax.scan(f, init, xs, length, reverse, unroll)
    else:
        xs_flat, xs_tree = jax.tree_util.tree_flatten(xs)
        carry = init
        ys = []
        maybe_reversed = reversed if reverse else lambda x: x
        for i in maybe_reversed(range(length)):
            xs_slice = [x[i] for x in xs_flat]
            carry, y = f(carry, jax.tree_util.tree_unflatten(xs_tree, xs_slice))
            ys.append(y)
        stacked_y = jax.tree_util.tree_map(lambda *y: onp.stack(y), *maybe_reversed(ys))
        return carry, stacked_y


def top_k(operand: ndarray, k: int) -> tuple[ndarray, ndarray]:
    """Returns top k values and their indices along the last axis of operand."""
    if which_np(operand) is jnp:
        return jax.lax.top_k(operand, k)
    else:
        top_ind = onp.argpartition(operand, -k)[-k:]
        sorted_ind = top_ind[onp.argsort(-operand[top_ind])]
        return operand[sorted_ind], sorted_ind


def while_loop(
    cond_fun: Callable[[X], Any], body_fun: Callable[[X], X], init_val: X
) -> X:
    """Call body_fun while cond_fun is true, starting with init_val."""
    if is_jitted():
        return jax.lax.while_loop(cond_fun, body_fun, init_val)
    else:
        val = init_val
        while cond_fun(val):
            val = body_fun(val)
        return val


def stop_gradient(x: X) -> X:
    """Returns x with zero gradient."""
    if which_np(x) is jnp:
        return jax.lax.stop_gradient(x)
    else:
        return x



"""Core implemented functions."""

from __future__ import annotations

from typing import Callable, Sequence, TypeVar

import numpy as onp

from jumpy import is_jax_installed
from jumpy.core import is_jitted, which_np
from jumpy.numpy import ndarray, take

try:
    import jax
    import jax.numpy as jnp
except ImportError:
    jax, jnp = None, None


F = TypeVar("F", bound=Callable)


def index_update(x: ndarray, idx: ndarray, y: ndarray) -> ndarray:
    """Pure equivalent of x[idx] = y."""
    if which_np(x, idx, y) is jnp:
        return jnp.array(x).at[idx].set(jnp.array(y))
    else:
        x = onp.copy(x)
        x[idx] = y
        return x


def vmap(fun: F, include: Sequence[bool] | None = None) -> F:
    """Creates a function which maps ``fun`` over argument axes."""
    if not is_jax_installed:
        raise NotImplementedError("This function requires the jax module")

    if is_jitted():
        in_axes = 0
        if include:
            in_axes = [0 if inc else None for inc in include]
        return jax.vmap(fun, in_axes=in_axes)

    def _batched(*args, include=include):
        if include is not None and len(include) != len(args):
            raise RuntimeError("Len of `args` list must match length of `include`.")

        # by default, vectorize over every arg
        if include is None:
            include = [True for _ in args]

        # determine number of parallel evaluations to unroll into serial evals
        batch_size = None
        for a, inc in zip(args, include):
            if inc:
                flat_args, _ = jax.tree_util.tree_flatten(a)
                batch_size = flat_args[0].shape[0]
                break

        # rebuild b_args for each serial evaluation
        rets = []
        for b_idx in range(batch_size):
            b_args = []
            for a, inc in zip(args, include):
                if inc:
                    b_args.append(take(a, b_idx))
                else:
                    b_args.append(a)
            rets.append(fun(*b_args))

        return jax.tree_util.tree_map(lambda *x: onp.stack(x), *rets)

    return _batched



"""Module for NumPy."""
from jumpy.numpy import linalg
from jumpy.numpy._factory_fns import (
    arange,
    array,
    atleast_1d,
    atleast_2d,
    atleast_3d,
    eye,
    linspace,
    meshgrid,
    ones,
    ones_like,
    reshape,
    zeros,
    zeros_like,
)
from jumpy.numpy._transform_data import (
    abs,
    all,
    amax,
    amin,
    any,
    arccos,
    arcsin,
    arctan2,
    arctanh,
    argmax,
    argmin,
    clip,
    concatenate,
    cos,
    cross,
    diag,
    dot,
    exp,
    expand_dims,
    floor,
    logical_and,
    logical_not,
    logical_or,
    matmul,
    maximum,
    mean,
    minimum,
    multiply,
    outer,
    repeat,
    roll,
    safe_arccos,
    safe_arcsin,
    sign,
    sin,
    sqrt,
    square,
    stack,
    sum,
    take,
    tanh,
    tile,
    var,
    where,
)
from jumpy.numpy._types import dtype, float32, inf, int32, ndarray, pi, uint8

__all__ = [
    # === primitives ===
    "dtype",
    "pi",
    "inf",
    "float32",
    "int32",
    "ndarray",
    "uint8",
    # === sub-modules ===
    "linalg",
    # === create data functions ===
    "array",
    "arange",
    "atleast_1d",
    "atleast_2d",
    "atleast_3d",
    "eye",
    "linspace",
    "meshgrid",
    "ones",
    "ones_like",
    "reshape",
    "zeros",
    "zeros_like",
    # === manipulation data functions ===
    "abs",
    "all",
    "amax",
    "amin",
    "any",
    "arccos",
    "arcsin",
    "arctan2",
    "arctanh",
    "argmax",
    "argmin",
    "clip",
    "concatenate",
    "cos",
    "cross",
    "diag",
    "dot",
    "exp",
    "expand_dims",
    "floor",
    "logical_and",
    "logical_not",
    "logical_or",
    "matmul",
    "maximum",
    "mean",
    "minimum",
    "multiply",
    "outer",
    "repeat",
    "roll",
    "safe_arccos",
    "safe_arcsin",
    "sign",
    "sin",
    "sqrt",
    "square",
    "stack",
    "sum",
    "take",
    "tanh",
    "tile",
    "var",
    "where",
]
